from __future__ import unicode_literals, print_function, absolute_import
from gevent.pywsgi import WSGIServer
from flask import (
    Flask, Blueprint, render_template, send_file, request, current_app, url_for,
)

import os
import json

from ..index import Index, PackageNotFound


api = Blueprint('api', __name__)


@api.route('/')
def api_index():
    return json.dumps(current_app.index.list_packages())


@api.route('/<package>')
def api_package(package):
    releases = []
    fmt = lambda u: url_for('download', package=u, _external=True)
    for release in current_app.index.package_releases(package, fmt):
        releases.append(release)
    if releases:
        return json.dumps(releases)
    else:
        return json.dumps({'status': 'error'}), 404


class Server(object):
    def __init__(self, args):
        self.args = args
        self.index = Index(args.curddir)
        self.index.scan()

        # Setting up the app
        self.app = Flask(__name__)
        self.app.index = self.index

        # Registering urls
        self.app.register_blueprint(api, url_prefix='/api')
        self.app.add_url_rule('/', 'index', self.web_index)
        self.app.add_url_rule('/s/<query>', 'search', self.web_search)
        self.app.add_url_rule('/p/<package>', 'download', self.web_download)
        self.app.add_url_rule('/p/<package>', 'upload', self.web_upload,
                              methods=('PUT',))

    def start(self):
        if self.args.debug:
            self.app.run(host=self.args.host, port=self.args.port, debug=True)
        else:
            WSGIServer((self.args.host, self.args.port), self.app).serve_forever()

    def web_index(self):
        return render_template('index.html', index=self.index)

    def web_search(self, query):
        try:
            path = self.index.get(query)
        except PackageNotFound:
            return 'package not found', 404
        except ValueError:
            return 'your query is wrong', 400
        return send_file(
            path, as_attachment=True,
            attachment_filename=os.path.basename(path))

    def web_download(self, package):
        with self.index.open(package) as pkg:
            return pkg.read()

    def web_upload(self, package):
        """
         * Smart enough to not save things we already have
         * Idempotent, you can call as many times as you need
         * The caller names the package (its basename)
        """
        pkg = request.files[package]
        self.index.from_data(package, pkg.read())
        return 'ok'



import threading


class Signal(list):
    pass


class SignalEmitter(object):

    def __init__(self):
        self.lock = threading.RLock()

    @property
    def name(self):
        return self.__class__.__name__.lower()

    def get_signal_or_explode(self, signal):
        try:
            with self.lock:
                return getattr(self, signal)
        except AttributeError:
            raise AttributeError(
                "There is no such signal ({0}) in this emitter ({1})".format(
                    signal, self.name))

    def connect(self, signal, callback):
        # Well, now we use the list-like interface of signal to file this new
        # callback under the previously retrieved signal container.
        self.get_signal_or_explode(signal).append(callback)

    def emit(self, signal, *args, **kwargs):
        for callback in self.get_signal_or_explode(signal):
            callback(*args, **kwargs)



from __future__ import absolute_import, unicode_literals, print_function
from collections import defaultdict
from distlib.util import parse_requirement
from .service import Service


def constraints(requirement):
    return (
        ','.join(' '.join(x) for x in requirement.constraints or ())
        or None)


class Maestro(object):

    def __init__(self, *args, **kwargs):
        super(Maestro, self).__init__(*args, **kwargs)
        self.mapping = defaultdict(dict)
        self.built = set()
        self.failed = set()

    def file_package(self, package, dependency_of=None):
        requirement = parse_requirement(package)
        version = constraints(requirement)
        self.mapping[requirement.name.lower()][version] = {
            'dependency_of': dependency_of,
            'data': None,
        }

    def _mark(self, attr, package, data):
        pkg = parse_requirement(package)
        name = pkg.name.lower()
        getattr(self, attr).add(name)
        self.mapping[name][constraints(pkg)]['data'] = data

    def mark_built(self, package, data):
        self._mark('built', package, data)

    def mark_failed(self, package, data):
        self._mark('failed', package, data)

    def get_data(self, package):
        requirement = parse_requirement(package)
        version = constraints(requirement)
        return self.mapping[requirement.name.lower()][version]['data']

    def best_version(self, package_name):
        versions = self.mapping[package_name].items()

        # We're looking for the version directly requested by the user. We
        # find it looking for versions that contain `None` in their field
        # `dependency_of`.
        for version, data in versions:
            if data['dependency_of'] is None:
                return version, data

        # There's no hard feelings about versions here. Meaning that the user
        # didn't request this package as a primary installation target.
        return versions[0]

    def should_queue(self, package):
        pkg = parse_requirement(package)
        return pkg.name.lower() not in self.mapping

    @property
    def pending_packages(self):
        return list(set(self.mapping.keys())
            .difference(self.built)
            .difference(self.failed))



from __future__ import unicode_literals, print_function

from . import Env
from .index import Index
from .util import expand_requirements, safe_name, AttrDict

import argparse
import os


DEFAULT_PYPI_INDEX_LIST = [
    'http://pypi.python.org/simple/',
]


class ValidationError(Exception):
    pass


def parse_args():
    parser = argparse.ArgumentParser(
        description='Curdles your cheesy code and extracts its binaries')

    parser.add_argument(
        '-r', '--requirements',
        help='A requirements file')

    parser.add_argument(
        '-i', '--index', action='append',
        help='PyPi compatible index url. Repeat as many times as you need')

    parser.add_argument(
        '-c', '--curdling-index', action='append', default=[],
        help='Curdling compatible index url. Repeat as many times as you need')

    parser.add_argument(
        '-u', '--upload', action='store_true', default=False,
        help='Upload your packages back to the curdling index')

    parser.add_argument(
        '-l', '--log-level', default=1, type=int,
        help=(
            'Increases the verbosity, goes from 0 (quiet) to '
            'the infinite and beyond (chatty)'))

    parser.add_argument(
        'packages', metavar='PKG', nargs='*',
        help='list of files to install')

    return parser.parse_args()


def prepare_args(args):
    if args.packages is None and args.requirements is None:
        raise ValidationError(
            'we need either at least one package or a requirements file')

    packages = [safe_name(req) for req in (args.packages or [])]
    if args.requirements:
        for pkg in expand_requirements(args.requirements):
            packages.append(pkg)

    return AttrDict(
        packages=packages,
        pypi_urls=args.index or DEFAULT_PYPI_INDEX_LIST,
        curdling_urls=args.curdling_index,
        upload=args.upload,
        log_level=args.log_level,
    )


def prepare_env():
    args = prepare_args(parse_args())

    # Setting up the index
    path = os.path.expanduser('~/.curds')
    index = Index(path)
    index.scan()

    # Configuration values for the environment
    args.update({
        'index': index,
        'concurrency': 10,
    })

    # Let's create the environment and start the required services
    env = Env(args)
    env.start_services()

    # Request the installation of the received package
    for pkg in args.packages:
        env.request_install('main', pkg)

    return env


def main():
    env = prepare_env()
    try:
        # All the installation requests were made, let's just wait here
        return env.run()
    except KeyboardInterrupt:
        print('\b\b')
        env.report()
        raise SystemExit(0)


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import absolute_import, unicode_literals, print_function
from .service import Service

import io
import os
import urllib3
import urlparse


class Uploader(Service):

    def __init__(self, *args, **kwargs):
        super(Uploader, self).__init__(*args, **kwargs)
        self.opener = urllib3.PoolManager()

    def handle(self, requester, package, sender_data):
        # Preparing the url to PUT the file
        path = sender_data.pop('path')
        server = sender_data.pop('server')
        package = os.path.basename(path)
        url = urlparse.urljoin(server, 'p/{0}'.format(package))

        # Sending the file to the server. Both `method` and `url` parameters
        # for calling `request_encode_body()` must be `str()` instances, not
        # unicode.
        contents = io.open(path, 'rb').read()
        self.opener.request_encode_body(
            b'PUT', bytes(url), {package: (package, contents)})
        return {'url': url}



from __future__ import unicode_literals, print_function
from curdling.web import Server

import argparse


def parse_args():
    parser = argparse.ArgumentParser(
        description='Share your cheese binaries with your folks')

    parser.add_argument(
        'curddir', metavar='DIRECTORY',
        help='Path for your cache directory')

    parser.add_argument(
        '-d', '--debug', action='store_true', default=False,
        help='Runs without gevent and enables debug')

    parser.add_argument(
        '-H', '--host', default='0.0.0.0',
        help='Host name to bind')

    parser.add_argument(
        '-p', '--port', type=int, default=8000,
        help='Port to bind')

    return parser.parse_args()


def main():
    server = Server(parse_args())
    server.start()


if __name__ == '__main__':
    main()



from __future__ import unicode_literals, print_function, absolute_import
from Queue import Queue
from .logging import ReportableError, Logger
from .signal import Signal, SignalEmitter

import threading
import time


class Service(SignalEmitter):

    def __init__(self, **args):
        super(Service, self).__init__()

        self.env = args.get('env')
        self.conf = args.pop('conf', {})
        self.index = args.pop('index', None)
        self.logger = Logger(self.name, args.get('log_level'))

        # Components to implement the thread pool
        self._queue = Queue()
        self.pool = []

        # Declaring signals
        self.started = Signal()
        self.finished = Signal()
        self.failed = Signal()

    def queue(self, requester, package, **data):
        self._queue.put((requester, package, data))
        self.logger.level(3, ' * queue(from=%s, to=%s, package=%s, data=%s)',
            requester, self.name, package, data)
        return self

    def start(self):
        self.logger.level(3, ' * %s.start()', self.name)
        for worker_num in range(self.conf.get('concurrency', 10)):
            worker = threading.Thread(target=self._worker)
            worker.daemon = True
            worker.start()
            self.pool.append(worker)
        return self

    def wait(self):
        while all(worker.is_alive() for worker in self.pool):
            time.sleep(1)

    def join(self):
        # We need to separate loops cause we can't actually tell which thread
        # got each sentinel
        for worker in self.pool:
            self._queue.put('STOP')  # Sending a sentinel
        for worker in self.pool:
            worker.join()
        self.workers = []

    # -- Private API --

    def _worker(self):
        for requester, package, sender_data in iter(self._queue.get, 'STOP'):
            self.logger.level(3, ' * %s[%s].run(package=%s, sender_data=%s)',
                self.name, threading.current_thread().name,
                package, sender_data)
            try:
                self.emit('started', self.name, package, **sender_data)
                data = self.handle(requester, package, sender_data)
                self._queue.task_done()
            except ReportableError as exc:
                self.emit('failed', self.name, package, path=exc)
                self.logger.level(0, " # %s.error(): %s", self.name, exc)
            except BaseException as exc:
                self.emit('failed', self.name, package, path=exc)
                self.logger.traceback(4,
                    'failed to run %s (requested by:%s) for package %s:',
                    self.name, requester, package)
            else:
                self.emit('finished', self.name, package, **(data or {}))
                self.logger.level(3, ' * %s.result(package=%s): %s ... ok',
                    self.name, package, data)



from __future__ import absolute_import, unicode_literals, print_function
from collections import defaultdict
from distlib.util import parse_requirement
from pkg_resources import parse_version, safe_name

from curdling.util import split_name, filehash

import os
import re
import shutil

FORMATS = ('whl', 'gz', 'bz', 'zip')

PKG_NAMES = [
    r'([\w\-]+)-([\d\.]+\d)[\.\-]',
    r'(\w+)-(.+)\.\w+$',
]


def pkg_name(name):
    for expr in PKG_NAMES:
        result = re.findall(expr, name)
        if result:
            return result[0]


def match_format(format_, name):
    ext = split_name(name)[1]
    if format_.startswith('~'):
        return format_[1:] != ext
    return format_ == ext


class PackageNotFound(Exception):
    def __init__(self, spec, formats):
        pkg = parse_requirement(spec)
        msg = ['The index does not have the requested package: ']
        msg.append(pkg.requirement)
        msg.append(formats and ' ({0})'.format(formats) or '')
        super(PackageNotFound, self).__init__(''.join(msg))


class Index(object):
    def __init__(self, base_path):
        self.base_path = base_path
        self.storage = defaultdict(lambda: defaultdict(list))

    def scan(self):
        if not os.path.isdir(self.base_path):
            return

        for file_name in os.listdir(self.base_path):
            destination = os.path.join(self.base_path, file_name)
            self.index(destination)

    def ensure_path(self, destination):
        path = os.path.dirname(destination)
        if not os.path.isdir(path):
            os.makedirs(path)
        return destination

    def index(self, path):
        pkg = os.path.basename(path)
        name, version = pkg_name(pkg)
        self.storage[safe_name(name.lower())][version].append(pkg)

    def from_file(self, path):
        # Moving the file around
        file_name = '.'.join(split_name(os.path.basename(path))[:2])
        destination = self.ensure_path(os.path.join(self.base_path, file_name))
        shutil.copy(path, destination)
        self.index(destination)
        return destination

    def from_data(self, path, data):
        # Build the name of the package based on its spec and extension
        file_name = '.'.join(split_name(os.path.basename(path))[:2])
        destination = self.ensure_path(os.path.join(self.base_path, file_name))
        with open(destination, 'wb') as fobj:
            fobj.write(data)
        self.index(destination)
        return destination

    def delete(self):
        shutil.rmtree(self.base_path)

    def list_packages(self):
        return self.storage.keys()

    def get_urlhash(self, url, fmt):
        """Returns the hash of the file of an internal url
        """
        with self.open(os.path.basename(url)) as f:
            return {'url': fmt(url), 'sha256': filehash(f, 'sha256')}

    def package_releases(self, package, url_fmt=lambda u: u):
        """List all versions of a package

        Along with the version, the caller also receives the file list with all
        the available formats.
        """
        return [{
            'name': package,
            'version': version,
            'urls': [self.get_urlhash(f, url_fmt) for f in files]
        } for version, files in self.storage.get(package, {}).items()]

    def open(self, fname, mode='r'):
        return open(os.path.abspath(os.path.join(
            self.base_path, os.path.basename(fname))), mode)

    def get(self, query):
        # Read both: "pkg==0.0.0" and "pkg==0.0.0,fmt"
        sym = ';'
        spec, format_ = (sym in query and (query.split(sym)) or (query, ''))
        requirement = parse_requirement(spec)

        # [First step] Looking up the package name parsed from the spec
        versions = self.storage.get(requirement.name.lower())
        if not versions:
            raise PackageNotFound(spec, format_)

        # [Second step] Filter out versions incompatible with our spec
        parsed_versions = {}
        [parsed_versions.update({parse_version(v): v}) for v in versions.keys()]

        filter_cmp = lambda x: all({
            '<':  lambda v: x <  parse_version(v),
            '<=': lambda v: x <= parse_version(v),
            '!=': lambda v: x != parse_version(v),
            '==': lambda v: x == parse_version(v),
            '>=': lambda v: x >= parse_version(v),
            '>':  lambda v: x >  parse_version(v),
        }[op](v) for op, v in requirement.constraints or [])

        compat_versions = filter(filter_cmp, parsed_versions.keys())
        if not compat_versions:
            raise PackageNotFound(spec, format_)

        # [Third step] Find best version to match the given format
        files = []

        # We don't have version or format, so we'll get the latest. Also,
        # we'll bring the wheels preferably, if they're available
        latest_version = versions[parsed_versions[max(compat_versions)]]
        if format_:
            files = filter(lambda n: match_format(format_, n), latest_version)
        else:
            wheels = filter(lambda n: match_format('whl', n), latest_version)
            files = wheels or latest_version

        # Unlucky, we really don't have those files
        if not files:
            raise PackageNotFound(spec, format_)

        # Yay, let's return the full path to the user
        return os.path.join(self.base_path, files[0])



from __future__ import unicode_literals, print_function, absolute_import
import traceback
import sys


class ReportableError(Exception):
    pass


class Logger(object):
    def __init__(self, name, run_level):
        self.name = name
        self.run_level = run_level or 200  # Will run with tests

    def level(self, level, msg, *args, **kwargs):
        indent = kwargs.get('indent')
        kind = kwargs.get('kind')

        if kind == 'traceback':
            indent = 4
        elif kind == 'error':
            indent = 2

        if indent:
            msg = '\n'.join(
                '{0}{1}'.format(' ' * indent, x)
                for x in msg.splitlines())

        if level <= self.run_level:
            print(msg % args, end=kwargs.get('end', '\n'))

    def traceback(self, level, msg, *args, **kwargs):
        # Print out banner
        exc = kwargs.get('exc')
        msg = msg if not exc else \
            "{0} ({1}: {2})".format(msg, exc.__class__.__name__, exc)
        self.level(level, msg, *args, kind='error')

        # Than the last traceback
        frames = traceback.extract_tb(sys.exc_info()[2])
        for frame in reversed(frames):
            self.level(level,
                ' %s:%s %s(): %s', *frame,
                indent=4, kind='traceback')




from __future__ import absolute_import, unicode_literals, print_function
from distlib.wheel import Wheel
from .service import Service
from .signal import Signal


class Dependencer(Service):

    def __init__(self, *args, **kwargs):
        super(Dependencer, self).__init__(*args, **kwargs)
        self.dependency_found = Signal()
        self.built = Signal()

    def handle(self, requester, package, sender_data):
        # Find the wheel
        path = sender_data.pop('path')
        wheel = Wheel(path)
        run_time_dependencies = wheel.metadata.requires_dist

        for spec in run_time_dependencies:
            # Packages might declare their "extras" here, so let's split it
            dependency, extra = (';' in spec and spec or spec + ';').split(';')
            self.emit('dependency_found', self.name,
                dependency, dependency_of=package)
        else:
            self.emit('built', self.name, package, path=path)

        return {'deps': run_time_dependencies}



from __future__ import absolute_import, unicode_literals, print_function
from distlib.util import parse_requirement

import io
import os
import re
import hashlib


INCLUDE_PATTERN = re.compile(r'-r\s*\b([^\b]+)')

LINK_PATTERN = re.compile(r'^([^\:]+):\/\/.+')


class AttrDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def split_name(fname):
    name, ext = os.path.splitext(fname)

    try:
        ext, frag = ext.split('#')
    except ValueError:
        frag = ''
    return name, ext[1:], frag


def safe_name(spec):
    return parse_requirement(spec).requirement


def expand_requirements(file_name):
    requirements = []

    for req in io.open(file_name).read().splitlines():
        req = req.split('#', 1)[0].strip()
        if not req:
            continue

        # Handling special lines that start with `-r`, so we can have files
        # including other files.
        include = INCLUDE_PATTERN.findall(req)
        if include:
            requirements.extend(expand_requirements(include[0]))
            continue

        # Handling links, let's do nothing with this guy right now
        link = LINK_PATTERN.findall(req)
        if link:
            continue

        # Finally, we're sure that it's just a package description
        requirements.append(safe_name(req))
    return requirements


def filehash(f, algo, block_size=2**20):
    algo = getattr(hashlib, algo)()
    while True:
        data = f.read(block_size)
        if not data:
            break
        algo.update(data)
    return algo.hexdigest()



from __future__ import absolute_import, unicode_literals, print_function
from .service import Service
from . import ReportableError

import os
import re
import sys
import subprocess
import shutil
import tempfile
import zipfile
import tarfile


# We'll use it to call the `setup.py` script of packages we're building
PYTHON_EXECUTABLE = sys.executable.encode(sys.getfilesystemencoding())

# Those are the formats we know how to extract, if you need to add a new one
# here, please refer to the page[0] to check the magic bits of the file type
# you wanna add.
#
# [0] http://www.garykessler.net/library/file_sigs.html
SUPPORTED_FORMATS = {
    b"\x1f\x8b\x08": "gz",
    b"\x42\x5a\x68": "bz2",
    b"\x50\x4b\x03\x04": "zip"
}

# Must be greater than the length of the biggest key of `SUPPORTED_FORMATS`, to
# be used as the block size to `file.read()` in `guess_file_type()`
SUPPORTED_FORMATS_MAX_LEN = max(len(x) for x in SUPPORTED_FORMATS)

# Matcher for egg-info directories
EGG_INFO_RE = re.compile(r'(-py\d\.\d)?\.egg-info', re.I)


def get_paths(directory='', check=False):
    paths = {}
    for sub in "purelib", "platlib", "headers", "data":
        path = os.path.join(directory, sub)
        if not check or os.path.exists(path):
            paths[sub] = path
    return paths


def guess_file_type(filename):
    with open(filename) as f:
        file_start = f.read(SUPPORTED_FORMATS_MAX_LEN)
    for magic, filetype in SUPPORTED_FORMATS.items():
        if file_start.startswith(magic):
            return filetype
    raise ReportableError('Unknown compress format for file %s' % filename)


class Script(object):

    def __init__(self, path):
        self.path = path

    def __call__(self, command, *custom_args):
        # What we're gonna run
        cwd = os.path.dirname(self.path)
        script = os.path.basename(self.path)

        # Building the argument list starting from the interpreter path. This
        # weird we're doing here was copied from `pip` and it basically forces
        # the usage of setuptools instead of distutils or any other weird
        # library people might be using.
        args = [PYTHON_EXECUTABLE]
        args.append('-c')
        args.append(
            r"import setuptools;__file__=%r;"
            r"exec(compile(open(__file__).read().replace('\r\n', '\n'), __file__, 'exec'))" % script)
        args.append(command)
        args.extend(custom_args)

        # Boom! Executing the command.
        null = open(os.devnull, 'w')
        subprocess.call(args, cwd=cwd, stdout=null, stderr=null)

        # Returning the path pointing to the generated file
        dist_dir = os.path.join(cwd, 'dist')
        return os.path.join(dist_dir, os.listdir(dist_dir)[0])


def unpack(package, destination):
    file_type = guess_file_type(package)

    # The only extensions we support currently
    if file_type == 'zip':
        fp = zipfile.ZipFile(package)
        get_names = fp.namelist
    elif file_type in ('gz', 'bz2'):
        fp = tarfile.open(package, 'r')
        get_names = lambda: [x.name for x in fp.getmembers()]

    # Find the setup.py script among the other contents
    try:
        setup_scripts = [x for x in get_names() if x.endswith('setup.py')]
        setup_py = sorted(setup_scripts, key=lambda e: len(e))[0]
        fp.extractall(destination)
    except ValueError:
        raise RuntimeError('No setup.py script was found here')
    finally:
        fp.close()
    return Script(os.path.join(destination, setup_py))


class Curdler(Service):

    def handle(self, requester, package, sender_data):
        source = sender_data.pop('path')

        # Place used to unpack the wheel
        destination = tempfile.mkdtemp()

        # Unpackaging the file we just received. The unpack function will give
        # us the path for the setup.py script and building the wheel file with
        # the `bdist_wheel` command.
        try:
            setup_py = unpack(package=source, destination=destination)
            wheel_file = setup_py('bdist_wheel')
            path = self.index.from_file(wheel_file)
        finally:
            shutil.rmtree(destination)

        # Finally, we just say where in the storage the file is
        return {'path': path}



import re
import os
from setuptools import setup, find_packages


def parse_requirements(path):
    """Rudimentary parser for the `requirements.txt` file

    We just want to separate regular packages from links to pass them to the
    `install_requires` and `dependency_links` params of the `setup()`
    function properly.
    """
    try:
        requirements = map(str.strip, local_file(path).splitlines())
    except IOError:
        raise RuntimeError("Couldn't find the `requirements.txt' file :(")

    links = []
    pkgs = []
    for req in requirements:
        if not req:
            continue
        if 'http:' in req or 'https:' in req:
            links.append(req)
            name, version = re.findall("\#egg=([^\-]+)-(.+$)", req)[0]
            pkgs.append('{0}=={1}'.format(name, version))
        else:
            pkgs.append(req)

    return pkgs, links


local_file = lambda f: \
    open(os.path.join(os.path.dirname(__file__), f)).read()

install_requires, dependency_links = \
    parse_requirements('requirements.txt')


if __name__ == '__main__':
    setup(
        name="curdling",
        version='0.2.0',
        description="curdling",
        long_description=local_file('README.md'),
        author='Lincoln Clarete',
        author_email='lincoln@clarete.li',
        url='https://github.com/clarete/curdling',
        packages=find_packages(exclude=['*tests*']),
        install_requires=install_requires,
        dependency_links=dependency_links,
        include_package_data=True,
        entry_points={
            'console_scripts': [
                'curdling = curdling.__main__:main',
                'curdling-server = curdling.web.__main__:main',
            ]
        },
        classifiers=[
            'Programming Language :: Python',
        ],
        extras_require={
            'server': parse_requirements('requirements-server.txt')[0],
        },
    )



from __future__ import unicode_literals, print_function, absolute_import
from collections import namedtuple
from functools import wraps
from distlib.database import DistributionPath
from distlib.util import parse_requirement

from .logging import Logger, ReportableError
from .index import PackageNotFound
from .maestro import Maestro

from .download import Downloader
from .wheelhouse import Curdler
from .dependency import Dependencer
from .installer import Installer
from .uploader import Uploader

import re
import time

SUCCESS = 0

FAILURE = 1

PACKAGE_BLACKLIST = (
    'setuptools',
)


def only(func, pattern):
    @wraps(func)
    def wrapper(requester, package, **data):
        if re.match(pattern, data.get('path', '')):
            return func(requester, package, **data)
    return wrapper


def mark(func):
    @wraps(func)
    def wrapper(requester, package, **data):
        return func(package, data['path'])
    return wrapper


class Env(object):
    def __init__(self, conf):
        self.conf = conf
        self.index = self.conf.get('index')
        self.logger = Logger('main', conf.get('log_level'))

    def start_services(self):
        # General params for all the services
        args = self.conf
        args.update({
            'env': self,
            'index': self.index,
            'conf': self.conf,
        })

        self.maestro = Maestro()
        self.downloader = Downloader(**args).start()
        self.curdler = Curdler(**args).start()
        self.dependencer = Dependencer(**args).start()

        # Building the pipeline
        self.downloader.connect('finished', only(self.curdler.queue, r'^(?!.*\.whl$)'))
        self.downloader.connect('finished', only(self.dependencer.queue, r'.*\.whl$'))
        self.downloader.connect('failed', mark(self.maestro.mark_failed))
        self.curdler.connect('finished', self.dependencer.queue)
        self.curdler.connect('failed', mark(self.maestro.mark_failed))
        self.dependencer.connect('dependency_found', self.request_install)
        self.dependencer.connect('built', mark(self.maestro.mark_built))

        # Not starting those guys since we don't actually have a lot to do here
        # right now. Check the `run` method, we'll call the installer and
        # uploader after making sure all the dependencies are installed.
        self.installer = Installer(**args)
        self.uploader = Uploader(**args)

    def report(self):
        if self.maestro.failed:
            self.logger.level(0, 'Some cheese was spilled in the process:')
        for package in self.maestro.failed:
            _, version = self.maestro.best_version(package)
            data = version.get('data')
            self.logger.level(0, " * %s: %s", data.__class__.__name__, data)

    def run(self):
        while self.maestro.pending_packages:
            time.sleep(0.5)

        if self.maestro.failed:
            self.report()
            return FAILURE

        # We've got everything we need, let's rock it off!
        self.run_installer()

        # Upload missing stuff that we couldn't find in curdling servers
        if self.conf.get('upload'):
            self.run_uploader()
        return SUCCESS

    def run_installer(self):
        self.installer.start()
        for package in self.maestro.mapping:
            _, version = self.maestro.best_version(package)
            self.installer.queue('main', package, path=version['data'])
        self.installer.join()

    def run_uploader(self):
        failures = self.downloader.get_servers_to_update()
        if not failures:
            return

        uploader = self.uploader.start()
        for server, packages in failures.items():
            for package in packages:
                _, data = self.maestro.best_version(package)
                uploader.queue('main', package,
                    path=data.get('data'), server=server)
        uploader.join()

    def check_installed(self, package):
        return DistributionPath().get_distribution(
            parse_requirement(package).name.replace('_', '-')) is not None

    def request_install(self, requester, package, **data):
        # If it's a blacklisted requirement, we should cowardly refuse to
        # install
        for blacklisted in PACKAGE_BLACKLIST:
            if package.startswith(blacklisted):
                self.logger.level(2,
                    "Cowardly refusing to install blacklisted "
                    "requirement `%s'", package)
                return False

        # Well, the package is installed, let's just bail
        if self.check_installed(package):
            return True

        # We shouldn't queue the same package twice
        if not self.maestro.should_queue(package):
            return False

        # Let's tell the maestro we have a new challenger
        self.maestro.file_package(package, dependency_of=data.get('dependency_of'))

        # Looking for built packages
        try:
            path = self.index.get("{0};whl".format(package))
            self.dependencer.queue(requester, package, path=path)
            return False
        except PackageNotFound:
            pass

        # Looking for downloaded packages. If there's packages of any of the
        # following distributions, we'll just build the wheel
        try:
            path = self.index.get("{0};~whl".format(package))
            self.curdler.queue(requester, package, path=path)
            return False
        except PackageNotFound:
            pass

        # Nops, we really don't have the package
        self.downloader.queue(requester, package, **data)
        return False

    def uninstall(self, package):
        from pip.commands.uninstall import UninstallCommand

        # We just overwrite the constructor here cause it's not actualy useful
        # unless you're creating another command, not calling as a library.
        class Uninstall(UninstallCommand):
            def __init__(self):
                pass

        # Just creating an object that pretends to be the option container for
        # the `run()` method.
        opts = namedtuple('Options', 'yes requirements')
        Uninstall().run(opts(yes=True, requirements=[]), [package])



from __future__ import absolute_import, unicode_literals, print_function
from distlib import database, metadata, compat, locators
from urlparse import urljoin

from . import util, ReportableError
from .service import Service
from .signal import Signal

import re
import json
import urllib3
import urllib3.exceptions
import distlib.version


def get_locator(conf):
    curds = [CurdlingLocator(u) for u in conf.get('curdling_urls', [])]
    pypi = [PyPiLocator(u) for u in conf.get('pypi_urls', [])]
    return AggregatingLocator(*(curds + pypi), scheme='legacy')


def find_packages(locator, package, versions):
    scheme = distlib.version.get_scheme(locator.scheme)
    matcher = scheme.matcher(package.requirement)

    result = {}
    if versions:
        slist = []
        for v in versions:
            if matcher.match(matcher.version_class(v)):
                slist.append(v)
        slist = sorted(slist, key=scheme.key)
        if len(slist):
            result = versions[slist[-1]]

    return result


class Pool(urllib3.PoolManager):

    def retrieve(self, url):
        attempts = 5
        headers = {}

        # Authentication
        parsed = compat.urlparse(url)
        if parsed.username:
            auth = '{0}:{1}'.format(parsed.username, parsed.password)
            headers = urllib3.util.make_headers(basic_auth=auth)

        # Params to be passed to request. The `preload_content` must be set to
        # False, otherwise `read()` wont honor `decode_content`.
        params = {'headers': headers, 'preload_content': False}

        # Request the url and ensure we've reached the final location
        response = self.request('GET', url, **params)
        return response, response.get_redirect_location() or url


class AggregatingLocator(locators.AggregatingLocator):

    def locate(self, requirement, prereleases=True):
        pkg = util.parse_requirement(requirement)
        for locator in self.locators:
            versions = locator.get_project(pkg.name)
            package = find_packages(locator, pkg, versions)
            if package:
                return package


class PyPiLocator(locators.SimpleScrapingLocator):

    def __init__(self, url, **kwargs):
        super(PyPiLocator, self).__init__(url, **kwargs)
        self.opener = Pool()

    def _get_project(self, name):
        return self._fetch(
            urljoin(self.base_url, '%s/' % compat.quote(name)),
            name)

    def _visit_link(self, project_name, link):
        self._seen.add(link)
        locators.logger.debug('_fetch() found link: %s', link)
        info = not self._is_platform_dependent(link) \
            and self.convert_url_to_download_info(link, project_name) \
            or None

        versions = {}
        if info:
            self._update_version_data(versions, info)
            return versions.items()[0]
        return None, None

    def _fetch(self, url, project_name, subvisit=False):
        locators.logger.debug('fetch(%s, %s)', url, project_name)
        versions = {}
        page = self.get_page(url)
        for link, rel in (page and page.links or []):
            # Let's instrospect one level down
            if self._should_queue(link, url, rel) and not subvisit:
                versions.update(self._fetch(link, project_name, subvisit=True))

            # Let's not see anything twice, I saw this check on distlib it
            # might be useful.
            if link not in self._seen:
                # Well, here we're ensuring that the first link of a given
                # version will be the one. Even if we find another package for
                # the same version, the first one will be used.
                version, distribution = self._visit_link(project_name, link)
                if version and version not in versions:
                    versions[version] = distribution
        return versions

    def get_page(self, url):
        # http://peak.telecommunity.com/DevCenter/EasyInstall#package-index-api
        scheme, netloc, path, _, _, _ = compat.urlparse(url)
        if scheme == 'file' and os.path.isdir(url2pathname(path)):
            url = urljoin(ensure_slash(url), 'index.html')

        # The `retrieve()` method follows any eventual redirects, so the
        # initial url might be different from the final one
        try:
            response, final_url = self.opener.retrieve(url)
        except urllib3.exceptions.MaxRetryError:
            return

        content_type = response.headers.get('content-type', '')
        if locators.HTML_CONTENT_TYPE.match(content_type):
            data = response.data
            encoding = response.headers.get('content-encoding')
            if encoding:
                decoder = self.decoders[encoding]   # fail if not found
                data = decoder(data)
            encoding = 'utf-8'
            m = locators.CHARSET.search(content_type)
            if m:
                encoding = m.group(1)
            try:
                data = data.decode(encoding)
            except UnicodeError:
                data = data.decode('latin-1')    # fallback
            return locators.Page(data, final_url)


class CurdlingLocator(locators.Locator):

    def __init__(self, url, **kwargs):
        super(CurdlingLocator, self).__init__(**kwargs)
        self.original_url = url
        self.url = url
        self.opener = Pool()
        self.packages_not_found = []

    def get_distribution_names(self):
        return json.loads(
            self.opener.retrieve(
                urljoin(self.url, 'api'))[0].data)

    def _get_project(self, name):
        # Retrieve the info
        url = urljoin(self.url, 'api/' + name)
        try:
            response, _ = self.opener.retrieve(url)
        except urllib3.exceptions.MaxRetryError:
            return None

        if response.status == 200:
            data = json.loads(response.data)
            return {v['version']: self._get_distribution(v) for v in data}
        else:
            self.packages_not_found.append(name)

    def _get_distribution(self, version):
        # Source url for the package
        source_url = version['urls'][0]  # TODO: prefer whl files

        # Build the metadata
        mdata = metadata.Metadata(scheme=self.scheme)
        mdata.name = version['name']
        mdata.version = version['version']
        mdata.source_url = mdata.download_url = source_url['url']

        # Building the dist and associating the download url
        distribution = database.Distribution(mdata)
        distribution.locator = self
        return distribution


class Downloader(Service):

    def __init__(self, *args, **kwargs):
        super(Downloader, self).__init__(*args, **kwargs)
        self.opener = Pool()
        self.locator = get_locator(self.conf)

    def handle(self, requester, package, sender_data):
        prereleases = self.conf.get('prereleases', True)
        requirement = self.locator.locate(package, prereleases)
        if requirement is None:
            raise ReportableError('Package `{0}\' not found'.format(package))
        return {"path": self.download(requirement.download_url)}

    def get_servers_to_update(self):
        failures = {}
        for locator in self.locator.locators:
            if isinstance(locator, CurdlingLocator) and locator.packages_not_found:
                failures[locator.original_url] = locator.packages_not_found
        return failures

    # -- Private API of the Download service --

    def download(self, url):
        response, _ = self.opener.retrieve(url)

        # Now that we're sure that our request was successful
        header = response.headers.get('content-disposition', '')
        file_name = re.findall(r'filename=([^;]+)', header)
        return self.index.from_data(
            file_name and file_name[0] or url, response.read(
                cache_content=True, decode_content=False))



from __future__ import absolute_import, unicode_literals, print_function
from distlib.wheel import Wheel
from distlib.util import parse_requirement
from .service import Service

import sys
import os.path


PREFIX = os.path.normpath(sys.prefix)


def get_distribution_paths(name):
    """Return target paths where the package content should be installed"""
    pyver = 'python' + sys.version[:3]

    paths = {
        'prefix' : '{prefix}',
        'data'   : '{prefix}/lib/{pyver}/site-packages',
        'purelib': '{prefix}/lib/{pyver}/site-packages',
        'platlib': '{prefix}/lib/{pyver}/site-packages',
        'headers': '{prefix}/include/{pyver}/{name}',
        'scripts': '{prefix}/bin',
    }

    # pip uses a similar path as an alternative to the system's (read-only)
    # include directory:
    if hasattr(sys, 'real_prefix'):  # virtualenv
        paths['headers'] = os.path.abspath(
            os.path.join(sys.prefix, 'include', 'site', pyver, name))

    # Replacing vars
    for key, val in paths.items():
        paths[key] = val.format(prefix=PREFIX, name=name, pyver=pyver)
    return paths


class Installer(Service):

    def handle(self, requester, package, sender_data):
        source = sender_data.pop('path')
        name = parse_requirement(package).name
        wheel = Wheel(source)
        wheel.install(get_distribution_paths(name))
